@@ -1,4 +1,4 @@
-// Copyright (c) 2018-2021  Brendan Molloy <brendan@bbqsrc.net>,
+// Copyright (c) 2018-2022  Brendan Molloy <brendan@bbqsrc.net>,
 //                          Ilya Solovyiov <ilya.solovyiov@gmail.com>,
 //                          Kai Ren <tyranron@gmail.com>
 //
@@ -22,12 +22,10 @@ use std::{
 
 use futures::{future::LocalBoxFuture, StreamExt as _};
 use regex::Regex;
-use structopt::{StructOpt, StructOptInternal};
 
 use crate::{
-    cli, event, parser, runner, step, tag::Ext as _, writer, Event,
-    FailureWriter, Parser, Runner, ScenarioType, Step, World, Writer,
-    WriterExt as _,
+    cli, event, parser, runner, step, tag::Ext as _, writer, Event, Parser,
+    Runner, ScenarioType, Step, World, Writer, WriterExt as _,
 };
 
 /// Top-level [Cucumber] executor.
@@ -55,7 +53,7 @@ where
     P: Parser<I>,
     R: Runner<W>,
     Wr: Writer<W>,
-    Cli: StructOpt,
+    Cli: clap::Args,
 {
     /// [`Parser`] sourcing [`Feature`]s for execution.
     ///
@@ -89,7 +87,7 @@ where
     P: Parser<I>,
     R: Runner<W>,
     Wr: Writer<W>,
-    Cli: StructOpt,
+    Cli: clap::Args,
 {
     /// Creates a custom [`Cucumber`] executor with the provided [`Parser`],
     /// [`Runner`] and [`Writer`].
@@ -195,14 +193,13 @@ where
     /// #     }
     /// # }
     /// #
-    /// # let fut = async {
+    /// # #[tokio::main(flavor = "current_thread")]
+    /// # async fn main() {
     /// MyWorld::cucumber()
     ///     .repeat_skipped()
     ///     .run_and_exit("tests/features/readme")
     ///     .await;
-    /// # };
-    /// #
-    /// # futures::executor::block_on(fut);
+    /// # }
     /// ```
     /// <script
     ///     id="asciicast-ox14HynkBIw8atpfhyfvKrsO3"
@@ -215,7 +212,10 @@ where
     #[must_use]
     pub fn repeat_skipped(
         self,
-    ) -> Cucumber<W, P, I, R, writer::Repeat<W, Wr>, Cli> {
+    ) -> Cucumber<W, P, I, R, writer::Repeat<W, Wr>, Cli>
+    where
+        Wr: writer::NonTransforming,
+    {
         Cucumber {
             parser: self.parser,
             runner: self.runner,
@@ -249,14 +249,13 @@ where
     /// #     }
     /// # }
     /// #
-    /// # let fut = async {
+    /// # #[tokio::main(flavor = "current_thread")]
+    /// # async fn main() {
     /// MyWorld::cucumber()
     ///     .fail_on_skipped()
     ///     .run_and_exit("tests/features/readme")
     ///     .await;
-    /// # };
-    /// #
-    /// # futures::executor::block_on(fut);
+    /// # }
     /// ```
     /// <script
     ///     id="asciicast-UcipuopO6IFEsIDty6vaJlCH9"
@@ -283,15 +282,14 @@ where
     /// #     }
     /// # }
     /// #
-    /// # let fut = async {
+    /// # #[tokio::main(flavor = "current_thread")]
+    /// # async fn main() {
     /// MyWorld::cucumber()
     ///     .repeat_failed()
     ///     .fail_on_skipped()
     ///     .run_and_exit("tests/features/readme")
     ///     .await;
-    /// # };
-    /// #
-    /// # futures::executor::block_on(fut);
+    /// # }
     /// ```
     /// <script
     ///     id="asciicast-ofOljvyEMb41OTLhE081QKv68"
@@ -299,20 +297,14 @@ where
     ///     async data-autoplay="true" data-rows="24">
     /// </script>
     ///
-    /// > ⚠️ __WARNING__: [`Cucumber::repeat_failed()`] should be called before
-    ///                   [`Cucumber::fail_on_skipped()`], as events pass from
-    ///                   outer [`Writer`]s to inner ones. So we need to
-    ///                   transform [`Skipped`] to [`Failed`] first, and only
-    ///                   then [`Repeat`] them.
-    ///
     /// [`Failed`]: crate::event::Step::Failed
-    /// [`Repeat`]: writer::Repeat
-    /// [`Scenario`]: gherkin::Scenario
-    /// [`Skipped`]: crate::event::Step::Skipped
     #[must_use]
     pub fn repeat_failed(
         self,
-    ) -> Cucumber<W, P, I, R, writer::Repeat<W, Wr>, Cli> {
+    ) -> Cucumber<W, P, I, R, writer::Repeat<W, Wr>, Cli>
+    where
+        Wr: writer::NonTransforming,
+    {
         Cucumber {
             parser: self.parser,
             runner: self.runner,
@@ -347,14 +339,13 @@ where
     /// #     }
     /// # }
     /// #
-    /// # let fut = async {
+    /// # #[tokio::main(flavor = "current_thread")]
+    /// # async fn main() {
     /// MyWorld::cucumber()
     ///     .fail_on_skipped()
     ///     .run_and_exit("tests/features/readme")
     ///     .await;
-    /// # };
-    /// #
-    /// # futures::executor::block_on(fut);
+    /// # }
     /// ```
     /// <script
     ///     id="asciicast-UcipuopO6IFEsIDty6vaJlCH9"
@@ -382,7 +373,8 @@ where
     /// #     }
     /// # }
     /// #
-    /// # let fut = async {
+    /// # #[tokio::main(flavor = "current_thread")]
+    /// # async fn main() {
     /// MyWorld::cucumber()
     ///     .repeat_if(|ev| {
     ///         use cucumber::event::{Cucumber, Feature, Rule, Scenario, Step};
@@ -409,9 +401,7 @@ where
     ///     .fail_on_skipped()
     ///     .run_and_exit("tests/features/readme")
     ///     .await;
-    /// # };
-    /// #
-    /// # futures::executor::block_on(fut);
+    /// # }
     /// ```
     /// <script
     ///     id="asciicast-ofOljvyEMb41OTLhE081QKv68"
@@ -419,16 +409,7 @@ where
     ///     async data-autoplay="true" data-rows="24">
     /// </script>
     ///
-    /// > ⚠️ __WARNING__: [`Cucumber::repeat_if()`] should be called before
-    ///                   [`Cucumber::fail_on_skipped()`], as events pass from
-    ///                   outer [`Writer`]s to inner ones. So we need to
-    ///                   transform [`Skipped`] to [`Failed`] first, and only
-    ///                   then [`Repeat`] them.
-    ///
     /// [`Failed`]: crate::event::Step::Failed
-    /// [`Repeat`]: writer::Repeat
-    /// [`Scenario`]: gherkin::Scenario
-    /// [`Skipped`]: crate::event::Step::Skipped
     #[must_use]
     pub fn repeat_if<F>(
         self,
@@ -436,6 +417,7 @@ where
     ) -> Cucumber<W, P, I, R, writer::Repeat<W, Wr, F>, Cli>
     where
         F: Fn(&parser::Result<Event<event::Cucumber<W>>>) -> bool,
+        Wr: writer::NonTransforming,
     {
         Cucumber {
             parser: self.parser,
@@ -454,10 +436,10 @@ where
     P: Parser<I>,
     R: Runner<W>,
     Wr: Writer<W> + for<'val> writer::Arbitrary<'val, W, String>,
-    Cli: StructOpt,
+    Cli: clap::Args,
 {
     /// Consider [`Skipped`] steps as [`Failed`] if their [`Scenario`] isn't
-    /// marked with `@allow_skipped` tag.
+    /// marked with `@allow.skipped` tag.
     ///
     /// It's useful option for ensuring that all the steps were covered.
     ///
@@ -489,14 +471,13 @@ where
     /// #     }
     /// # }
     /// #
-    /// # let fut = async {
+    /// # #[tokio::main(flavor = "current_thread")]
+    /// # async fn main() {
     /// MyWorld::cucumber()
     ///     .fail_on_skipped()
     ///     .run_and_exit("tests/features/readme")
     ///     .await;
-    /// # };
-    /// #
-    /// # futures::executor::block_on(fut);
+    /// # }
     /// ```
     /// <script
     ///     id="asciicast-IHLxMEgku9BtBVkR4k2DtOjMd"
@@ -505,7 +486,7 @@ where
     /// </script>
     ///
     /// To intentionally suppress some [`Skipped`] steps failing, use the
-    /// `@allow_skipped` tag:
+    /// `@allow.skipped` tag:
     /// ```gherkin
     /// Feature: Animal feature
     ///
@@ -514,7 +495,7 @@ where
     ///     When I feed the cat
     ///     Then the cat is not hungry
     ///
-    ///   @allow_skipped
+    ///   @allow.skipped
     ///   Scenario: If we feed a satiated dog it will not become hungry
     ///     Given a satiated dog
     ///     When I feed the dog
@@ -570,14 +551,13 @@ where
     /// #     }
     /// # }
     /// #
-    /// # let fut = async {
+    /// # #[tokio::main(flavor = "current_thread")]
+    /// # async fn main() {
     /// MyWorld::cucumber()
     ///     .fail_on_skipped_with(|_, _, s| !s.tags.iter().any(|t| t == "dog"))
     ///     .run_and_exit("tests/features/readme")
     ///     .await;
-    /// # };
-    /// #
-    /// # futures::executor::block_on(fut);
+    /// # }
     /// ```
     /// ```gherkin
     /// Feature: Animal feature
@@ -645,8 +625,8 @@ where
     W: World,
     P: Parser<I>,
     R: Runner<W>,
-    Wr: Writer<W>,
-    Cli: StructOpt + StructOptInternal,
+    Wr: Writer<W> + writer::Normalized,
+    Cli: clap::Args,
 {
     /// Runs [`Cucumber`].
     ///
@@ -664,7 +644,7 @@ where
     /// using them inside [`Cucumber`].
     ///
     /// Also, any additional custom CLI options may be specified as a
-    /// [`StructOpt`] deriving type, used as the last type parameter of
+    /// [`clap::Args`] deriving type, used as the last type parameter of
     /// [`cli::Opts`].
     ///
     /// > ⚠️ __WARNING__: Any CLI options of [`Parser`], [`Runner`], [`Writer`]
@@ -679,7 +659,6 @@ where
     /// # use async_trait::async_trait;
     /// # use cucumber::{cli, WorldInit};
     /// # use futures::FutureExt as _;
-    /// # use structopt::StructOpt;
     /// # use tokio::time;
     /// #
     /// # #[derive(Debug, WorldInit)]
@@ -694,18 +673,19 @@ where
     /// #     }
     /// # }
     /// #
-    /// # let fut = async {
-    /// #[derive(StructOpt)]
+    /// # #[tokio::main(flavor = "current_thread")]
+    /// # async fn main() {
+    /// #[derive(clap::Args)]
     /// struct CustomCli {
     ///     /// Additional time to wait in a before hook.
-    ///     #[structopt(
+    ///     #[clap(
     ///         long,
     ///         parse(try_from_str = humantime::parse_duration)
     ///     )]
     ///     before_time: Option<Duration>,
     /// }
     ///
-    /// let cli = cli::Opts::<_, _, _, CustomCli>::from_args();
+    /// let cli = cli::Opts::<_, _, _, CustomCli>::parsed();
     /// let time = cli.custom.before_time.unwrap_or_default();
     ///
     /// MyWorld::cucumber()
@@ -713,13 +693,7 @@ where
     ///     .with_cli(cli)
     ///     .run_and_exit("tests/features/readme")
     ///     .await;
-    /// # };
-    /// #
-    /// # tokio::runtime::Builder::new_current_thread()
-    /// #    .enable_all()
-    /// #    .build()
-    /// #    .unwrap()
-    /// #    .block_on(fut);
+    /// # }
     /// ```
     /// ```gherkin
     /// Feature: Animal feature
@@ -743,7 +717,7 @@ where
         cli: cli::Opts<P::Cli, R::Cli, Wr::Cli, CustomCli>,
     ) -> Cucumber<W, P, I, R, Wr, CustomCli>
     where
-        CustomCli: StructOpt,
+        CustomCli: clap::Args,
     {
         let Cucumber {
             parser,
@@ -787,15 +761,14 @@ where
     /// #     }
     /// # }
     /// #
-    /// # let fut = async {
+    /// # #[tokio::main(flavor = "current_thread")]
+    /// # async fn main() {
     /// MyWorld::cucumber()
     ///     .filter_run("tests/features/readme", |_, _, sc| {
     ///         sc.tags.iter().any(|t| t == "cat")
     ///     })
     ///     .await;
-    /// # };
-    /// #
-    /// # futures::executor::block_on(fut);
+    /// # }
     /// ```
     /// ```gherkin
     /// Feature: Animal feature
@@ -836,18 +809,31 @@ where
             runner: runner_cli,
             writer: writer_cli,
             ..
-        } = self.cli.unwrap_or_else(cli::Opts::<_, _, _, _>::from_args);
+        } = self.cli.unwrap_or_else(cli::Opts::<_, _, _, _>::parsed);
 
-        let filter = move |f: &gherkin::Feature,
-                           r: Option<&gherkin::Rule>,
-                           s: &gherkin::Scenario| {
+        let filter = move |feat: &gherkin::Feature,
+                           rule: Option<&gherkin::Rule>,
+                           scenario: &gherkin::Scenario| {
             re_filter.as_ref().map_or_else(
                 || {
-                    tags_filter
-                        .as_ref()
-                        .map_or_else(|| filter(f, r, s), |f| f.eval(&s.tags))
+                    tags_filter.as_ref().map_or_else(
+                        || filter(feat, rule, scenario),
+                        |tags| {
+                            // The order `Feature` -> `Rule` -> `Scenario`
+                            // matters here.
+                            tags.eval(
+                                feat.tags
+                                    .iter()
+                                    .chain(
+                                        rule.into_iter()
+                                            .flat_map(|r| r.tags.iter()),
+                                    )
+                                    .chain(scenario.tags.iter()),
+                            )
+                        },
+                    )
                 },
-                |f| f.is_match(&s.name),
+                |re| re.is_match(&scenario.name),
             )
         };
 
@@ -862,16 +848,16 @@ where
 
         let filtered = features.map(move |feature| {
             let mut feature = feature?;
-            let scenarios = mem::take(&mut feature.scenarios);
-            feature.scenarios = scenarios
+            let feat_scenarios = mem::take(&mut feature.scenarios);
+            feature.scenarios = feat_scenarios
                 .into_iter()
                 .filter(|s| filter(&feature, None, s))
                 .collect();
 
             let mut rules = mem::take(&mut feature.rules);
             for r in &mut rules {
-                let scenarios = mem::take(&mut r.scenarios);
-                r.scenarios = scenarios
+                let rule_scenarios = mem::take(&mut r.scenarios);
+                r.scenarios = rule_scenarios
                     .into_iter()
                     .filter(|s| filter(&feature, Some(r), s))
                     .collect();
@@ -896,7 +882,7 @@ where
     P: Debug + Parser<I>,
     R: Debug + Runner<W>,
     Wr: Debug + Writer<W>,
-    Cli: StructOpt,
+    Cli: clap::Args,
 {
     fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
         f.debug_struct("Cucumber")
@@ -913,7 +899,7 @@ pub(crate) type DefaultCucumber<W, I> = Cucumber<
     parser::Basic,
     I,
     runner::Basic<W>,
-    writer::Summarized<writer::Normalized<W, writer::Basic>>,
+    writer::Summarize<writer::Normalize<W, writer::Basic>>,
 >;
 
 impl<W, I> Default for DefaultCucumber<W, I>
@@ -922,10 +908,10 @@ where
     I: AsRef<Path>,
 {
     fn default() -> Self {
-        Cucumber::custom(
+        Self::custom(
             parser::Basic::new(),
             runner::Basic::default(),
-            writer::Basic::new().normalized().summarized(),
+            writer::Basic::stdout().summarized(),
         )
     }
 }
@@ -944,20 +930,20 @@ where
     ///     `@serial` [tag] is present on a [`Scenario`];
     ///   * Allowed to run up to 64 [`Concurrent`] [`Scenario`]s.
     ///
-    /// * [`Writer`] — [`Normalized`] and [`Summarized`] [`writer::Basic`].
+    /// * [`Writer`] — [`Normalize`] and [`Summarize`] [`writer::Basic`].
     ///
     /// [`Concurrent`]: runner::basic::ScenarioType::Concurrent
-    /// [`Normalized`]: writer::Normalized
+    /// [`Normalize`]: writer::Normalize
     /// [`Parser`]: parser::Parser
     /// [`Scenario`]: gherkin::Scenario
     /// [`Serial`]: runner::basic::ScenarioType::Serial
     /// [`ScenarioType`]: runner::basic::ScenarioType
-    /// [`Summarized`]: writer::Summarized
+    /// [`Summarize`]: writer::Summarize
     ///
     /// [tag]: https://cucumber.io/docs/cucumber/api/#tags
     #[must_use]
     pub fn new() -> Self {
-        Cucumber::default()
+        Self::default()
     }
 }
 
@@ -966,7 +952,7 @@ where
     W: World,
     R: Runner<W>,
     Wr: Writer<W>,
-    Cli: StructOpt,
+    Cli: clap::Args,
     I: AsRef<Path>,
 {
     /// Sets the provided language of [`gherkin`] files.
@@ -989,7 +975,7 @@ where
     W: World,
     P: Parser<I>,
     Wr: Writer<W>,
-    Cli: StructOpt,
+    Cli: clap::Args,
     F: Fn(
             &gherkin::Feature,
             Option<&gherkin::Rule>,
@@ -1187,8 +1173,8 @@ where
     W: World,
     P: Parser<I>,
     R: Runner<W>,
-    Wr: FailureWriter<W>,
-    Cli: StructOpt + StructOptInternal,
+    Wr: writer::Failure<W> + writer::Normalized,
+    Cli: clap::Args,
 {
     /// Runs [`Cucumber`].
     ///
@@ -1238,15 +1224,14 @@ where
     /// #     }
     /// # }
     /// #
-    /// # let fut = async {
+    /// # #[tokio::main(flavor = "current_thread")]
+    /// # async fn main() {
     /// MyWorld::cucumber()
     ///     .filter_run_and_exit("tests/features/readme", |_, _, sc| {
     ///         sc.tags.iter().any(|t| t == "cat")
     ///     })
     ///     .await;
-    /// # };
-    /// #
-    /// # futures::executor::block_on(fut);
+    /// # }
     /// ```
     /// ```gherkin
     /// Feature: Animal feature
