@@ -1,4 +1,4 @@
-// Copyright (c) 2018-2021  Brendan Molloy <brendan@bbqsrc.net>,
+// Copyright (c) 2018-2022  Brendan Molloy <brendan@bbqsrc.net>,
 //                          Ilya Solovyiov <ilya.solovyiov@gmail.com>,
 //                          Kai Ren <tyranron@gmail.com>
 //
@@ -126,68 +126,68 @@ fn expand_scenario(
     /// [`Examples`]: gherkin::Examples
     #[allow(clippy::expect_used)]
     static TEMPLATE_REGEX: Lazy<Regex> =
-        Lazy::new(|| Regex::new(r"<(\S+)>").expect("incorrect Regex"));
+        Lazy::new(|| Regex::new(r"<([^>\s]+)>").expect("incorrect Regex"));
 
-    let (header, vals) = match scenario
+    if scenario.examples.is_empty() {
+        return vec![Ok(scenario)];
+    }
+
+    scenario
         .examples
-        .as_ref()
-        .and_then(|ex| ex.table.rows.split_first())
-    {
-        Some(s) => s,
-        None => return vec![Ok(scenario)],
-    };
+        .iter()
+        .filter_map(|ex| ex.table.rows.split_first().map(|(h, v)| (h, v, ex)))
+        .flat_map(|(header, vals, example)| {
+            vals.iter()
+                .map(|v| header.iter().zip(v))
+                .enumerate()
+                .zip(iter::repeat((example.position, example.tags.iter())))
+        })
+        .map(|((id, row), (position, tags))| {
+            let replace_templates = |str: &str, pos| {
+                let mut err = None;
+                let replaced = TEMPLATE_REGEX
+                    .replace_all(str, |cap: &regex::Captures<'_>| {
+                        let name = cap.get(1).unwrap().as_str();
 
-    let table = vals.iter().map(|v| header.iter().zip(v));
-    table
-        .enumerate()
-        .map(|(id, row)| {
-            let mut modified = scenario.clone();
+                        row.clone()
+                            .find_map(|(k, v)| (name == k).then(|| v.as_str()))
+                            .unwrap_or_else(|| {
+                                err = Some(ExpandExamplesError {
+                                    pos,
+                                    name: name.to_owned(),
+                                    path: path.cloned(),
+                                });
+                                ""
+                            })
+                    })
+                    .into_owned();
+
+                err.map_or_else(|| Ok(replaced), Err)
+            };
+
+            let mut expanded = scenario.clone();
 
             // This is done to differentiate `Hash`es of
             // scenario outlines with the same examples.
-            modified.position = scenario
-                .examples
-                .as_ref()
-                .map_or_else(|| scenario.position, |ex| ex.position);
-            modified.position.line += id + 1;
+            expanded.position = position;
+            expanded.position.line += id + 1;
 
-            let mut err = None;
+            expanded.tags.extend(tags.cloned());
 
-            for s in &mut modified.steps {
-                let pos = s.position;
-                let to_replace = iter::once(&mut s.value).chain(
-                    s.table.iter_mut().flat_map(|t| {
+            expanded.name =
+                replace_templates(&expanded.name, expanded.position)?;
+            for s in &mut expanded.steps {
+                for value in iter::once(&mut s.value)
+                    .chain(s.docstring.iter_mut())
+                    .chain(s.table.iter_mut().flat_map(|t| {
                         t.rows.iter_mut().flat_map(|r| r.iter_mut())
-                    }),
-                );
-
-                for value in to_replace {
-                    *value = TEMPLATE_REGEX
-                        .replace_all(value, |c: &regex::Captures<'_>| {
-                            let name = c.get(1).unwrap().as_str();
-
-                            row.clone()
-                                .find_map(|(k, v)| {
-                                    (name == k).then(|| v.as_str())
-                                })
-                                .unwrap_or_else(|| {
-                                    err = Some(ExpandExamplesError {
-                                        pos,
-                                        name: name.to_owned(),
-                                        path: path.cloned(),
-                                    });
-                                    ""
-                                })
-                        })
-                        .into_owned();
-                }
-
-                if let Some(e) = err {
-                    return Err(e);
+                    }))
+                {
+                    *value = replace_templates(value, s.position)?;
                 }
             }
 
-            Ok(modified)
+            Ok(expanded)
         })
         .collect()
 }
