@@ -1,4 +1,4 @@
-// Copyright (c) 2018-2021  Brendan Molloy <brendan@bbqsrc.net>,
+// Copyright (c) 2018-2022  Brendan Molloy <brendan@bbqsrc.net>,
 //                          Ilya Solovyiov <ilya.solovyiov@gmail.com>,
 //                          Kai Ren <tyranron@gmail.com>
 //
@@ -33,6 +33,7 @@
     clippy::decimal_literal_representation,
     clippy::else_if_without_else,
     clippy::empty_line_after_outer_attr,
+    clippy::equatable_if_let,
     clippy::exit,
     clippy::expect_used,
     clippy::fallible_impl_from,
@@ -59,6 +60,7 @@
     clippy::rc_buffer,
     clippy::rc_mutex,
     clippy::rest_pat_in_fully_bound_structs,
+    clippy::same_name_method,
     clippy::shadow_unrelated,
     clippy::str_to_string,
     clippy::string_add,
@@ -118,10 +120,10 @@ pub use gherkin;
 
 #[cfg(feature = "macros")]
 #[doc(inline)]
-pub use self::codegen::WorldInit;
+pub use self::codegen::{Parameter, WorldInit};
 #[cfg(feature = "macros")]
 #[doc(inline)]
-pub use cucumber_codegen::{given, then, when, WorldInit};
+pub use cucumber_codegen::{given, then, when, Parameter, WorldInit};
 
 #[doc(inline)]
 pub use self::{
