@@ -15,10 +15,10 @@ use std::{
     collections::HashMap,
     convert::identity,
     fmt, mem,
+    ops::ControlFlow,
     panic::{self, AssertUnwindSafe},
-    path::PathBuf,
     sync::{
-        atomic::{AtomicBool, AtomicUsize, Ordering},
+        atomic::{AtomicBool, Ordering},
         Arc,
     },
 };
@@ -48,6 +48,10 @@ pub struct Cli {
     /// value configured in tests runner, or 64 by default.
     #[clap(long, short, name = "int")]
     pub concurrency: Option<usize>,
+
+    /// Run tests until the first failure.
+    #[clap(long)]
+    pub fail_fast: bool,
 }
 
 /// Type determining whether [`Scenario`]s should run concurrently or
@@ -101,6 +105,11 @@ pub type AfterHookFn<World> = for<'a> fn(
     Option<&'a mut World>,
 ) -> LocalBoxFuture<'a, ()>;
 
+/// Alias for a failed [`Scenario`].
+///
+/// [`Scenario`]: gherkin::Scenario
+type Failed = bool;
+
 /// Default [`Runner`] implementation which follows [_order guarantees_][1] from
 /// the [`Runner`] trait docs.
 ///
@@ -148,6 +157,9 @@ pub struct Basic<
     /// [`Scenario`]: gherkin::Scenario
     /// [`Step`]: gherkin::Step
     after_hook: Option<After>,
+
+    /// Indicates whether execution should be stopped after the first failure.
+    fail_fast: bool,
 }
 
 // Implemented manually to omit redundant trait bounds on `World` and to omit
@@ -171,6 +183,7 @@ impl<World> Basic<World, ()> {
             which_scenario: (),
             before_hook: None,
             after_hook: None,
+            fail_fast: false,
         }
     }
 }
@@ -192,6 +205,7 @@ impl<World> Default for Basic<World> {
             which_scenario,
             before_hook: None,
             after_hook: None,
+            fail_fast: false,
         }
     }
 }
@@ -210,6 +224,18 @@ impl<World, Which, Before, After> Basic<World, Which, Before, After> {
         self
     }
 
+    /// Run tests until the first failure.
+    ///
+    /// __NOTE__: All the already started [`Scenario`]s at the moment of failure
+    ///           will be finished.
+    ///
+    /// [`Scenario`]: gherkin::Scenario
+    #[must_use]
+    pub const fn fail_fast(mut self) -> Self {
+        self.fail_fast = true;
+        self
+    }
+
     /// Function determining whether a [`Scenario`] is [`Concurrent`] or
     /// a [`Serial`] one.
     ///
@@ -231,6 +257,7 @@ impl<World, Which, Before, After> Basic<World, Which, Before, After> {
             steps,
             before_hook,
             after_hook,
+            fail_fast,
             ..
         } = self;
         Basic {
@@ -239,6 +266,7 @@ impl<World, Which, Before, After> Basic<World, Which, Before, After> {
             which_scenario: func,
             before_hook,
             after_hook,
+            fail_fast,
         }
     }
 
@@ -263,6 +291,7 @@ impl<World, Which, Before, After> Basic<World, Which, Before, After> {
             steps,
             which_scenario,
             after_hook,
+            fail_fast,
             ..
         } = self;
         Basic {
@@ -271,6 +300,7 @@ impl<World, Which, Before, After> Basic<World, Which, Before, After> {
             which_scenario,
             before_hook: Some(func),
             after_hook,
+            fail_fast,
         }
     }
 
@@ -304,6 +334,7 @@ impl<World, Which, Before, After> Basic<World, Which, Before, After> {
             steps,
             which_scenario,
             before_hook,
+            fail_fast,
             ..
         } = self;
         Basic {
@@ -312,6 +343,7 @@ impl<World, Which, Before, After> Basic<World, Which, Before, After> {
             which_scenario,
             before_hook,
             after_hook: Some(func),
+            fail_fast,
         }
     }
 
@@ -392,8 +424,10 @@ where
             which_scenario,
             before_hook,
             after_hook,
+            fail_fast,
         } = self;
 
+        let fail_fast = if cli.fail_fast { true } else { fail_fast };
         let buffer = Features::default();
         let (sender, receiver) = mpsc::unbounded();
 
@@ -402,6 +436,7 @@ where
             features,
             which_scenario,
             sender.clone(),
+            fail_fast,
         );
         let execute = execute(
             buffer,
@@ -410,6 +445,7 @@ where
             sender,
             before_hook,
             after_hook,
+            fail_fast,
         );
 
         stream::select(
@@ -436,6 +472,7 @@ async fn insert_features<W, S, F>(
     features: S,
     which_scenario: F,
     sender: mpsc::UnboundedSender<parser::Result<Event<event::Cucumber<W>>>>,
+    fail_fast: bool,
 ) where
     S: Stream<Item = parser::Result<gherkin::Feature>> + 'static,
     F: Fn(
@@ -452,7 +489,7 @@ async fn insert_features<W, S, F>(
             // If the receiver end is dropped, then no one listens for events
             // so we can just stop from here.
             Err(e) => {
-                if sender.unbounded_send(Err(e)).is_err() {
+                if sender.unbounded_send(Err(e)).is_err() || fail_fast {
                     break;
                 }
             }
@@ -464,14 +501,25 @@ async fn insert_features<W, S, F>(
 
 /// Retrieves [`Feature`]s and executes them.
 ///
+/// # Events
+///
+/// - [`Scenario`] events are emitted by [`Executor`].
+/// - If [`Scenario`] was first or last for particular [`Rule`] or [`Feature`],
+///   emits starting or finishing events for them.
+///
 /// [`Feature`]: gherkin::Feature
+/// [`Rule`]: gherkin::Rule
+/// [`Scenario`]: gherkin::Scenario
 async fn execute<W, Before, After>(
     features: Features,
     max_concurrent_scenarios: Option<usize>,
     collection: step::Collection<W>,
-    sender: mpsc::UnboundedSender<parser::Result<Event<event::Cucumber<W>>>>,
+    event_sender: mpsc::UnboundedSender<
+        parser::Result<Event<event::Cucumber<W>>>,
+    >,
     before_hook: Option<Before>,
     after_hook: Option<After>,
+    fail_fast: bool,
 ) where
     W: World,
     Before: 'static
@@ -500,60 +548,86 @@ async fn execute<W, Before, After>(
     let hook = panic::take_hook();
     panic::set_hook(Box::new(|_| {}));
 
-    let mut executor =
-        Executor::new(collection, before_hook, after_hook, sender);
+    let (finished_sender, finished_receiver) = mpsc::unbounded();
+    let mut storage = FinishedRulesAndFeatures::new(finished_receiver);
+    let executor = Executor::new(
+        collection,
+        before_hook,
+        after_hook,
+        event_sender,
+        finished_sender,
+    );
 
-    executor.send(event::Cucumber::Started);
+    executor.send_event(event::Cucumber::Started);
 
+    // TODO: Replace with `ControlFlow::map_break()` once stabilized:
+    //       https://github.com/rust-lang/rust/issues/75744
+    let map_break = |cf| match cf {
+        ControlFlow::Continue(cont) => cont,
+        ControlFlow::Break(()) => Some(0),
+    };
+
+    let mut started_scenarios = ControlFlow::Continue(max_concurrent_scenarios);
+    let mut run_scenarios = stream::FuturesUnordered::new();
     loop {
-        let runnable = features.get(max_concurrent_scenarios).await;
-        if runnable.is_empty() {
+        let runnable = features.get(map_break(started_scenarios)).await;
+        if run_scenarios.is_empty() && runnable.is_empty() {
             if features.is_finished() {
                 break;
             }
             continue;
         }
 
-        let started = executor.start_scenarios(&runnable);
-        executor.send_all(started);
+        let started = storage.start_scenarios(&runnable);
+        executor.send_all_events(started);
 
-        drop(
-            runnable
-                .into_iter()
-                .map(|(f, r, s)| executor.run_scenario(f, r, s))
-                .collect::<future::JoinAll<_>>()
-                .await,
-        );
+        if let ControlFlow::Continue(Some(sc)) = &mut started_scenarios {
+            *sc -= runnable.len();
+        }
 
-        executor.cleanup_finished_rules_and_features();
+        for (f, r, s) in runnable {
+            run_scenarios.push(executor.run_scenario(f, r, s));
+        }
+
+        if run_scenarios.next().await.is_some() {
+            if let ControlFlow::Continue(Some(sc)) = &mut started_scenarios {
+                *sc += 1;
+            }
+        }
+
+        while let Ok(Some((feat, rule, scenario_failed))) =
+            storage.finished_receiver.try_next()
+        {
+            if let Some(r) = rule {
+                if let Some(f) =
+                    storage.rule_scenario_finished(Arc::clone(&feat), r)
+                {
+                    executor.send_event(f);
+                }
+            }
+            if let Some(f) = storage.feature_scenario_finished(feat) {
+                executor.send_event(f);
+            }
+
+            if fail_fast && scenario_failed {
+                started_scenarios = ControlFlow::Break(());
+            }
+        }
     }
 
-    executor.send(event::Cucumber::Finished);
+    // This is done in case of `fail_fast: true`, when not all `Scenario`s might
+    // be executed.
+    executor.send_all_events(storage.finish_all_rules_and_features());
+
+    executor.send_event(event::Cucumber::Finished);
 
     panic::set_hook(hook);
 }
 
-/// Stores currently ran [`Feature`]s and notifies about their state of
-/// completion.
+/// Runs [`Scenario`]s and notifies about their state of completion.
 ///
-/// [`Feature`]: gherkin::Feature.
+/// [`Scenario`]: gherkin::Scenario
 struct Executor<W, Before, After> {
-    /// Number of finished [`Scenario`]s of [`Feature`].
-    ///
-    /// [`Feature`]: gherkin::Feature
-    /// [`Scenario`]: gherkin::Scenario
-    features_scenarios_count: HashMap<Arc<gherkin::Feature>, AtomicUsize>,
-
-    /// Number of finished [`Scenario`]s of [`Rule`].
-    ///
-    /// We also store path to `.feature` file so [`Rule`]s with same names and
-    /// spans in different files will have different hashes.
-    ///
-    /// [`Rule`]: gherkin::Rule
-    /// [`Scenario`]: gherkin::Scenario
-    rule_scenarios_count:
-        HashMap<(Option<PathBuf>, Arc<gherkin::Rule>), AtomicUsize>,
-
     /// [`Step`]s [`Collection`].
     ///
     /// [`Collection`]: step::Collection
@@ -574,10 +648,21 @@ struct Executor<W, Before, After> {
     /// [`Step`]: gherkin::Step
     after_hook: Option<After>,
 
-    /// Sender for notifying state of [`Feature`]s completion.
+    /// Sender for [`Scenario`] [events][1].
     ///
-    /// [`Feature`]: gherkin::Feature
-    sender: mpsc::UnboundedSender<parser::Result<Event<event::Cucumber<W>>>>,
+    /// [`Scenario`]: gherkin::Scenario
+    /// [1]: event::Scenario
+    event_sender:
+        mpsc::UnboundedSender<parser::Result<Event<event::Cucumber<W>>>>,
+
+    /// Sender for notifying of [`Scenario`]s completion.
+    ///
+    /// [`Scenario`]: gherkin::Scenario
+    finished_sender: mpsc::UnboundedSender<(
+        Arc<gherkin::Feature>,
+        Option<Arc<gherkin::Rule>>,
+        Failed,
+    )>,
 }
 
 impl<W: World, Before, After> Executor<W, Before, After>
@@ -602,17 +687,21 @@ where
         collection: step::Collection<W>,
         before_hook: Option<Before>,
         after_hook: Option<After>,
-        sender: mpsc::UnboundedSender<
+        event_sender: mpsc::UnboundedSender<
             parser::Result<Event<event::Cucumber<W>>>,
         >,
+        finished_sender: mpsc::UnboundedSender<(
+            Arc<gherkin::Feature>,
+            Option<Arc<gherkin::Rule>>,
+            Failed,
+        )>,
     ) -> Self {
         Self {
-            features_scenarios_count: HashMap::new(),
-            rule_scenarios_count: HashMap::new(),
             collection,
             before_hook,
             after_hook,
-            sender,
+            event_sender,
+            finished_sender,
         }
     }
 
@@ -621,8 +710,6 @@ where
     /// # Events
     ///
     /// - Emits all [`Scenario`] events.
-    /// - If [`Scenario`] was last for particular [`Rule`] or [`Feature`], also
-    ///   emits finishing events for them.
     ///
     /// [`Feature`]: gherkin::Feature
     /// [`Rule`]: gherkin::Rule
@@ -672,13 +759,14 @@ where
             event::Scenario::step_failed,
         );
 
-        self.send(event::Cucumber::scenario(
+        self.send_event(event::Cucumber::scenario(
             Arc::clone(&feature),
             rule.clone(),
             Arc::clone(&scenario),
             event::Scenario::Started,
         ));
 
+        let mut is_failed = false;
         let world = async {
             let before_hook = self
                 .run_before_hook(&feature, rule.as_ref(), &scenario)
@@ -725,34 +813,33 @@ where
                 })
                 .await
         }
+        .inspect_err(|e| {
+            if e.is_none() {
+                is_failed = true;
+            }
+        })
         .await
         .unwrap_or_else(identity);
 
         self.run_after_hook(world, &feature, rule.as_ref(), &scenario)
             .await
-            .map_or((), drop);
+            .map_or_else(|_| is_failed = true, drop);
 
-        self.send(event::Cucumber::scenario(
+        self.send_event(event::Cucumber::scenario(
             Arc::clone(&feature),
             rule.clone(),
             Arc::clone(&scenario),
             event::Scenario::Finished,
         ));
 
-        if let Some(r) = rule {
-            if let Some(f) =
-                self.rule_scenario_finished(Arc::clone(&feature), r)
-            {
-                self.send(f);
-            }
-        }
-
-        if let Some(f) = self.feature_scenario_finished(feature) {
-            self.send(f);
-        }
+        self.scenario_finished(feature, rule, is_failed);
     }
 
     /// Executes [`HookType::Before`], if present.
+    ///
+    /// # Events
+    ///
+    /// - Emits [`HookType::Before`] event.
     async fn run_before_hook(
         &self,
         feature: &Arc<gherkin::Feature>,
@@ -776,7 +863,7 @@ where
         };
 
         if let Some(hook) = self.before_hook.as_ref() {
-            self.send(event::Cucumber::scenario(
+            self.send_event(event::Cucumber::scenario(
                 Arc::clone(feature),
                 rule.map(Arc::clone),
                 Arc::clone(scenario),
@@ -798,7 +885,7 @@ where
 
             match fut.await {
                 Ok(world) => {
-                    self.send(event::Cucumber::scenario(
+                    self.send_event(event::Cucumber::scenario(
                         Arc::clone(feature),
                         rule.map(Arc::clone),
                         Arc::clone(scenario),
@@ -807,7 +894,7 @@ where
                     Ok(Some(world))
                 }
                 Err((info, world)) => {
-                    self.send(event::Cucumber::scenario(
+                    self.send_event(event::Cucumber::scenario(
                         Arc::clone(feature),
                         rule.map(Arc::clone),
                         Arc::clone(scenario),
@@ -826,6 +913,10 @@ where
     }
 
     /// Executes [`HookType::After`], if present.
+    ///
+    /// # Event
+    ///
+    /// - Emits [`HookType::After`] event.
     async fn run_after_hook(
         &self,
         mut world: Option<W>,
@@ -834,7 +925,7 @@ where
         scenario: &Arc<gherkin::Scenario>,
     ) -> Result<Option<W>, ()> {
         if let Some(hook) = self.after_hook.as_ref() {
-            self.send(event::Cucumber::scenario(
+            self.send_event(event::Cucumber::scenario(
                 Arc::clone(feature),
                 rule.map(Arc::clone),
                 Arc::clone(scenario),
@@ -857,7 +948,7 @@ where
             #[allow(clippy::shadow_unrelated)]
             match fut.await {
                 Ok(world) => {
-                    self.send(event::Cucumber::scenario(
+                    self.send_event(event::Cucumber::scenario(
                         Arc::clone(feature),
                         rule.map(Arc::clone),
                         Arc::clone(scenario),
@@ -866,7 +957,7 @@ where
                     Ok(world)
                 }
                 Err((info, world)) => {
-                    self.send(event::Cucumber::scenario(
+                    self.send_event(event::Cucumber::scenario(
                         Arc::clone(feature),
                         rule.map(Arc::clone),
                         Arc::clone(scenario),
@@ -908,7 +999,7 @@ where
             event::StepError,
         ) -> event::Cucumber<W>,
     {
-        self.send(started(Arc::clone(&step)));
+        self.send_event(started(Arc::clone(&step)));
 
         let run = async {
             let (step_fn, captures, ctx) = match self.collection.find(&step) {
@@ -953,39 +1044,139 @@ where
         #[allow(clippy::shadow_unrelated)]
         match run.await {
             Ok((Some(captures), Some(world))) => {
-                self.send(passed(step, captures));
+                self.send_event(passed(step, captures));
                 Ok(world)
             }
             Ok((_, world)) => {
-                self.send(skipped(step));
+                self.send_event(skipped(step));
                 Err(world)
             }
             Err((err, captures, world)) => {
-                self.send(failed(step, captures, world.map(Arc::new), err));
+                self.send_event(failed(
+                    step,
+                    captures,
+                    world.map(Arc::new),
+                    err,
+                ));
                 Err(None)
             }
         }
     }
 
+    /// Notifies [`FinishedRulesAndFeatures`] about [`Scenario`] being finished.
+    ///
+    /// [`Scenario`]: gherkin::Scenario
+    fn scenario_finished(
+        &self,
+        feature: Arc<gherkin::Feature>,
+        rule: Option<Arc<gherkin::Rule>>,
+        is_failed: Failed,
+    ) {
+        // If the receiver end is dropped, then no one listens for events
+        // so we can just ignore it.
+        drop(
+            self.finished_sender
+                .unbounded_send((feature, rule, is_failed)),
+        );
+    }
+
+    /// Notifies with the given [`Cucumber`] event.
+    ///
+    /// [`Cucumber`]: event::Cucumber
+    fn send_event(&self, event: event::Cucumber<W>) {
+        // If the receiver end is dropped, then no one listens for events
+        // so we can just ignore it.
+        drop(self.event_sender.unbounded_send(Ok(Event::new(event))));
+    }
+
+    /// Notifies with the given [`Cucumber`] events.
+    ///
+    /// [`Cucumber`]: event::Cucumber
+    fn send_all_events(
+        &self,
+        events: impl Iterator<Item = event::Cucumber<W>>,
+    ) {
+        for v in events {
+            // If the receiver end is dropped, then no one listens for events
+            // so we can just stop from here.
+            if self.event_sender.unbounded_send(Ok(Event::new(v))).is_err() {
+                break;
+            }
+        }
+    }
+}
+
+/// Stores currently running [`Rule`]s and [`Feature`]s and notifies about their
+/// state of completion.
+///
+/// [`Feature`]: gherkin::Feature
+/// [`Rule`]: gherkin::Rule
+struct FinishedRulesAndFeatures {
+    /// Number of finished [`Scenario`]s of [`Feature`].
+    ///
+    /// [`Feature`]: gherkin::Feature
+    /// [`Scenario`]: gherkin::Scenario
+    features_scenarios_count: HashMap<Arc<gherkin::Feature>, usize>,
+
+    /// Number of finished [`Scenario`]s of [`Rule`].
+    ///
+    /// We also store path to a [`Feature`], so [`Rule`]s with same names and
+    /// spans in different `.feature` files will have different hashes.
+    ///
+    /// [`Feature`]: gherkin::Feature
+    /// [`Rule`]: gherkin::Rule
+    /// [`Scenario`]: gherkin::Scenario
+    rule_scenarios_count:
+        HashMap<(Arc<gherkin::Feature>, Arc<gherkin::Rule>), usize>,
+
+    /// Receiver for notifying state of [`Scenario`]s completion.
+    ///
+    /// [`Scenario`]: gherkin::Scenario
+    finished_receiver: mpsc::UnboundedReceiver<(
+        Arc<gherkin::Feature>,
+        Option<Arc<gherkin::Rule>>,
+        Failed,
+    )>,
+}
+
+impl FinishedRulesAndFeatures {
+    /// Creates a new [`FinishedRulesAndFeatures`] store.
+    fn new(
+        finished_receiver: mpsc::UnboundedReceiver<(
+            Arc<gherkin::Feature>,
+            Option<Arc<gherkin::Rule>>,
+            Failed,
+        )>,
+    ) -> Self {
+        Self {
+            features_scenarios_count: HashMap::new(),
+            rule_scenarios_count: HashMap::new(),
+            finished_receiver,
+        }
+    }
+
     /// Marks [`Rule`]'s [`Scenario`] as finished and returns [`Rule::Finished`]
     /// event if no [`Scenario`]s left.
     ///
     /// [`Rule`]: gherkin::Rule
     /// [`Rule::Finished`]: event::Rule::Finished
     /// [`Scenario`]: gherkin::Scenario
-    fn rule_scenario_finished(
-        &self,
+    fn rule_scenario_finished<W>(
+        &mut self,
         feature: Arc<gherkin::Feature>,
         rule: Arc<gherkin::Rule>,
     ) -> Option<event::Cucumber<W>> {
         let finished_scenarios = self
             .rule_scenarios_count
-            .get(&(feature.path.clone(), Arc::clone(&rule)))
-            .unwrap_or_else(|| panic!("No Rule {}", rule.name))
-            .fetch_add(1, Ordering::SeqCst)
-            + 1;
-        (rule.scenarios.len() == finished_scenarios)
-            .then(|| event::Cucumber::rule_finished(feature, rule))
+            .get_mut(&(Arc::clone(&feature), Arc::clone(&rule)))
+            .unwrap_or_else(|| panic!("No Rule {}", rule.name));
+        *finished_scenarios += 1;
+        (rule.scenarios.len() == *finished_scenarios).then(|| {
+            let _ = self
+                .rule_scenarios_count
+                .remove(&(Arc::clone(&feature), Arc::clone(&rule)));
+            event::Cucumber::rule_finished(feature, rule)
+        })
     }
 
     /// Marks [`Feature`]'s [`Scenario`] as finished and returns
@@ -994,19 +1185,38 @@ where
     /// [`Feature`]: gherkin::Feature
     /// [`Feature::Finished`]: event::Feature::Finished
     /// [`Scenario`]: gherkin::Scenario
-    fn feature_scenario_finished(
-        &self,
+    fn feature_scenario_finished<W>(
+        &mut self,
         feature: Arc<gherkin::Feature>,
     ) -> Option<event::Cucumber<W>> {
         let finished_scenarios = self
             .features_scenarios_count
-            .get(&feature)
-            .unwrap_or_else(|| panic!("No Feature {}", feature.name))
-            .fetch_add(1, Ordering::SeqCst)
-            + 1;
+            .get_mut(&feature)
+            .unwrap_or_else(|| panic!("No Feature {}", feature.name));
+        *finished_scenarios += 1;
         let scenarios = feature.count_scenarios();
-        (scenarios == finished_scenarios)
-            .then(|| event::Cucumber::feature_finished(feature))
+        (scenarios == *finished_scenarios).then(|| {
+            let _ = self.features_scenarios_count.remove(&feature);
+            event::Cucumber::feature_finished(feature)
+        })
+    }
+
+    /// Marks all the unfinished [`Rule`]s and [`Feature`]s as finished, and
+    /// returns all the appropriate finished events.
+    ///
+    /// [`Feature`]: gherkin::Feature
+    /// [`Rule`]: gherkin::Rule
+    fn finish_all_rules_and_features<W>(
+        &mut self,
+    ) -> impl Iterator<Item = event::Cucumber<W>> + '_ {
+        self.rule_scenarios_count
+            .drain()
+            .map(|((feat, rule), _)| event::Cucumber::rule_finished(feat, rule))
+            .chain(
+                self.features_scenarios_count
+                    .drain()
+                    .map(|(feat, _)| event::Cucumber::feature_finished(feat)),
+            )
     }
 
     /// Marks [`Scenario`]s as started and returns [`Rule::Started`] and
@@ -1018,7 +1228,7 @@ where
     /// [`Rule`]: gherkin::Rule
     /// [`Rule::Started`]: event::Rule::Started
     /// [`Scenario`]: gherkin::Scenario
-    fn start_scenarios(
+    fn start_scenarios<W>(
         &mut self,
         runnable: impl AsRef<
             [(
@@ -1037,7 +1247,7 @@ where
                 .entry(Arc::clone(&feature))
                 .or_insert_with(|| {
                     started_features.push(feature);
-                    0.into()
+                    0
                 });
         }
 
@@ -1051,10 +1261,10 @@ where
         {
             let _ = self
                 .rule_scenarios_count
-                .entry((feat.path.clone(), Arc::clone(&rule)))
+                .entry((Arc::clone(&feat), Arc::clone(&rule)))
                 .or_insert_with(|| {
                     started_rules.push((feat, rule));
-                    0.into()
+                    0
                 });
         }
 
@@ -1067,51 +1277,6 @@ where
                     .map(|(f, r)| event::Cucumber::rule_started(f, r)),
             )
     }
-
-    /// Removes all finished [`Rule`]s and [`Feature`]s as all their events are
-    /// emitted already.
-    ///
-    /// [`Feature`]: gherkin::Feature
-    /// [`Rule`]: gherkin::Rule
-    fn cleanup_finished_rules_and_features(&mut self) {
-        self.features_scenarios_count = self
-            .features_scenarios_count
-            .drain()
-            .filter(|(f, count)| {
-                f.count_scenarios() != count.load(Ordering::SeqCst)
-            })
-            .collect();
-
-        self.rule_scenarios_count = self
-            .rule_scenarios_count
-            .drain()
-            .filter(|((_, r), count)| {
-                r.scenarios.len() != count.load(Ordering::SeqCst)
-            })
-            .collect();
-    }
-
-    /// Notifies with the given [`Cucumber`] event.
-    ///
-    /// [`Cucumber`]: event::Cucumber
-    fn send(&self, event: event::Cucumber<W>) {
-        // If the receiver end is dropped, then no one listens for events
-        // so we can just ignore it.
-        drop(self.sender.unbounded_send(Ok(Event::new(event))));
-    }
-
-    /// Notifies with the given [`Cucumber`] events.
-    ///
-    /// [`Cucumber`]: event::Cucumber
-    fn send_all(&self, events: impl Iterator<Item = event::Cucumber<W>>) {
-        for v in events {
-            // If the receiver end is dropped, then no one listens for events
-            // so we can just stop from here.
-            if self.sender.unbounded_send(Ok(Event::new(v))).is_err() {
-                break;
-            }
-        }
-    }
 }
 
 /// [`Scenario`]s storage.
