@@ -1,4 +1,4 @@
-// Copyright (c) 2018-2021  Brendan Molloy <brendan@bbqsrc.net>,
+// Copyright (c) 2018-2022  Brendan Molloy <brendan@bbqsrc.net>,
 //                          Ilya Solovyiov <ilya.solovyiov@gmail.com>,
 //                          Kai Ren <tyranron@gmail.com>
 //
@@ -19,7 +19,6 @@ use std::{
     fmt,
     hash::{Hash, Hasher},
     iter,
-    path::PathBuf,
 };
 
 use derive_more::{Deref, DerefMut, Display, Error};
@@ -162,7 +161,7 @@ impl<World> Collection<World> {
             StepType::Then => &self.then,
         };
 
-        let mut matches = collection
+        let mut captures = collection
             .iter()
             .filter_map(|((re, loc), step_fn)| {
                 let mut captures = re.capture_locations();
@@ -171,15 +170,15 @@ impl<World> Collection<World> {
             })
             .collect::<Vec<_>>();
 
-        let (_, _, whole_match, captures, step_fn) = match matches.len() {
+        let (_, _, whole_match, captures, step_fn) = match captures.len() {
             0 => return Ok(None),
             // Instead of `.unwrap()` to avoid documenting `# Panics` section.
-            1 => matches.pop().unwrap_or_else(|| unreachable!()),
+            1 => captures.pop().unwrap_or_else(|| unreachable!()),
             _ => {
                 return Err(AmbiguousMatchError {
-                    possible_matches: matches
+                    possible_matches: captures
                         .into_iter()
-                        .map(|(re, loc, ..)| (re.clone(), loc.clone()))
+                        .map(|(re, loc, ..)| (re.clone(), *loc))
                         .collect(),
                 })
             }
@@ -229,10 +228,10 @@ pub struct AmbiguousMatchError {
 }
 
 /// Location of a [`Step`] [`fn`] automatically filled by a proc macro.
-#[derive(Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+#[derive(Clone, Copy, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
 pub struct Location {
     /// Path to the file where [`Step`] [`fn`] is located.
-    pub path: PathBuf,
+    pub path: &'static str,
 
     /// Line of the file where [`Step`] [`fn`] is located.
     pub line: u32,
